/* dot_product.c by Robert Marmorstein 
 * Solution to Problem 3.9 in "Introduction to Parallel Programming" by Peter S. Pachecho
 * Feb. 12, 2021
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include <mpi.h>


int N; 				//Size of the vectors, specified on the command line
int local_n; 			//Size of vector for each process (Global, since used in every function)

int my_rank, comm_sz;		//Process ID and number of processes (Global so all functions can use them)

void Read_Data(double* local_A, double* local_B, double *C)
{
	double *A; //First vector
	double *B; //Second vector

	if (my_rank == 0) {
		printf("Please type in %d numbers for each vector, followed by a single scalar:\n", N);

		//Allocate memory to store the vectors
		A = (double *)calloc(N, sizeof(double));
		B = (double *)calloc(N, sizeof(double));

		//Read first vector
		for (int i=0; i < N; ++i) {
			scanf("%lf", &A[i]);
		}

		//Read second vector
		for (int i=0; i < N; ++i) {
			scanf("%lf", &B[i]);
		}

		//Read scalar
		scanf("%lf", C);
	}

	//Send vector data to each process
	MPI_Scatter(A, local_n, MPI_DOUBLE, local_A, local_n, MPI_DOUBLE, 0, MPI_COMM_WORLD);	
	MPI_Scatter(B, local_n, MPI_DOUBLE, local_B, local_n, MPI_DOUBLE, 0, MPI_COMM_WORLD);	

	//Send scalar
	MPI_Bcast(C, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);

	//Free global vectors now that we're done with them
	if (my_rank == 0) {
		free(A);
		free(B);
	}
}

double Dot_Product(double* local_A, double* local_B)
{
	double total = 0.0;
	double local_product = 0.0;

	//Compute the dot product of the elements THIS process is responsible for
	for (int i=0; i < local_n; ++i) {
		local_product += local_A[i]*local_B[i];	
	}

	//Use Reduce to combine results
	MPI_Reduce(&local_product, &total, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);

	return total;
}

void Vector_Product(double* local_A, double* local_B, double C, double* vector_product)
{
	double *local_vector;
	local_vector = (double *)calloc(local_n, sizeof(double));
	for (int i=0; i < local_n; ++i) {
		local_vector[i] = C * local_A[i] * local_B[i];	
	}
	MPI_Gather(local_vector, local_n, MPI_DOUBLE, vector_product, local_n, MPI_DOUBLE, 0, MPI_COMM_WORLD);
	free(local_vector);
}

int main(int argc, char* argv[])
{
	double C;  //Scalar

	double dot_product;	//Result of (A dot B)
	double *vector_product;	//Result of A * B * C

	double* local_A;	//Each process has its local vector which contains SOME of the elements of each vector
	double* local_B;

	if (argc < 2) {
		printf("Syntax: mpiexec -np 4 --host amdahl:2,babbage:2 ./DotProduct <N>\n");
		exit(1);
	}

	MPI_Init(NULL, NULL);
	
	//Read in our rank and the number of processes
	MPI_Comm_size(MPI_COMM_WORLD, &comm_sz);
	MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);

	//Get N from the command line arguments
	N = atoi(argv[1]);
	
	//Compute per process number of vector elements
	local_n = N / comm_sz; //Assumes N is evenly divisible by comm_sz.

	//Allocate memory for local copies of each vector
	local_A = (double *)calloc(local_n, sizeof(double));
	local_B = (double *)calloc(local_n, sizeof(double));

	//Read in the data using process 0, then distribute to other processes
	Read_Data(local_A, local_B, &C);

	//Do the dot product
	dot_product = Dot_Product(local_A, local_B);
	if (my_rank == 0) {
		printf("Dot Product: %lf\n", dot_product);
	}

	//Do the scaled vector product

	//Set aside memory for the result 
	if (my_rank == 0) {
		vector_product = (double *) calloc(N, sizeof(double));
	}

	Vector_Product(local_A, local_B, C, vector_product);
	
	if (my_rank == 0) {	
		for (int i = 0; i < N; ++i)
		{
			printf("%lf ", vector_product[i]);
		}
		printf("\n");
	}

	if (my_rank == 0) {
		free(vector_product);
	}

	MPI_Finalize();
}
